/*
 * gDMM Digital Multimeter Application
 *
 * File: dmmread.c
 *
 * Copyright (c) 2003 Johannes Heimansberg
 *
 * gDMM is a GTK+ 2 based application which reads data from a Digital Multimeter
 * (Metex/Voltcraft or compatible, e.g. M-3650D) through the serial interface
 * (RS-232) and draws a graph out of this data. Measurements can also be 
 * saved to/loaded from disk.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <termio.h>
#include <math.h>
#include <malloc.h>
#include "dmmread.h"
#define STORE_MAX 128000
#define DMM_DATA_LENGTH_MAX 32
#define TRUE 1
#define FALSE 0
/* #define VERBOSE 1 */

/* Returns the value which was decoded last */
double dmm_get_last_value(DMM *dmm)
{
	return dmm->value;
}

/* Returns the maximum value */
double dmm_get_max_value(DMM *dmm)
{
	return (*dmm).value_max;
}

/* Returns the minimum value */
double dmm_get_min_value(DMM *dmm)
{
	return dmm->value_min;
}

/* Returns the number of stored values */
int dmm_get_number_of_stored_values(DMM *dmm)
{
	return dmm->value_array_counter;
}

/* Returns the stored value with index "number" as a float */
double dmm_get_stored_value(DMM *dmm, int number)
{
	if (number > dmm_get_number_of_stored_values(dmm) - 1)
		number = dmm_get_number_of_stored_values(dmm) - 1;
	return dmm_decode_string_m3650(dmm, number);
}

/* Returns the rawdata which where read last */
char* dmm_get_rawdata(DMM *dmm)
{
	return dmm->rawdata;
}

/* Returns the unit of the measurement */
char* dmm_get_unit(DMM *dmm)
{
	return dmm->unit;
}

/* Returns TRUE if the unit has changed during the measurement */
int dmm_is_unit_changed(DMM *dmm)
{
	return dmm->unit_changed;
}

/* (Re-)initializes the DMM struct */
void dmm_init_struct(DMM *dmm)
{
	dmm_clear_struct(dmm);

	strcpy(dmm->rawdata, "\0");
	strcpy(dmm->unit, "\0");
	dmm->value = 0;
	dmm->value_max = 0;
	dmm->value_min = 0;
	dmm->unit_changed = FALSE;
}

/* Free the memory that was allocated for storing the measured values */
void dmm_clear_struct(DMM *dmm)
{
	int i;
#ifdef VERBOSE
	printf("dmmread: Freeing memory: %d strings.\n", dmm->value_array_counter);
#endif
	for (i = 0; i < dmm->value_array_counter; i++)
		free(dmm->rawdata_array[i]);
	dmm->value_array_counter = 0;
}

/* Opens and initializes the serial port */
int dmm_open_port(char* serial_device, int baudrate, int data_bits, int stop_bits)
{
	int            fd; /* File descriptor for the port */
	int            bps = 0;
	struct termios options;
	int            mdlns;

	fd = open(serial_device, O_RDWR | O_NOCTTY);
	if (fd < 0)
	{
		perror("dmmread: dmm_open_port: Unable to open serial device - ");
	}
    else
		fcntl(fd, F_SETFL, 0);

  	options.c_oflag = 0;
	options.c_lflag = 0;
	options.c_iflag = IGNBRK | IGNPAR | INPCK;

    options.c_cflag = CLOCAL | CREAD;

	switch (data_bits)
	{
		case 5:
			options.c_cflag |= CS5; /* 5 Data bits */
			break;
		case 6:
			options.c_cflag |= CS6; /* 6 Data bits */
			break;
		case 7:
			options.c_cflag |= CS7; /* 7 Data bits */
			break;
		case 8:
			options.c_cflag |= CS8; /* 8 Data bits */
	}

	if (stop_bits == 2) options.c_cflag |=  CSTOPB;  /* 2 Stop bits */

#ifdef VERBOSE
	printf("dmmread: Serial port settings:\n");
	printf("dmmread:\tParity: %d\n", ((options.c_cflag & PARENB) ? 1 : 0));
	printf("dmmread:\t2 Stop bits: %d\n ", ((options.c_cflag & CSTOPB) ? 1 : 0));
	printf("dmmread:\t5 Data bits: %d\n ", ((options.c_cflag & CS5) ? 1 : 0));
	printf("dmmread:\t6 Data bits: %d\n ", ((options.c_cflag & CS6) ? 1 : 0));
	printf("dmmread:\t7 Data bits: %d\n ", ((options.c_cflag & CS7) ? 1 : 0));
	printf("dmmread:\t8 Data bits: %d\n ", ((options.c_cflag & CS8) ? 1 : 0));
	printf("dmmread:\tHW flow control: %d\n ", ((options.c_cflag & CRTSCTS) ? 1 : 0));
#endif

	options.c_cc[VTIME]= 2;
	options.c_cc[VMIN]= 0;

#ifdef VERBOSE
	printf("dmmread: Set baud rate to %dbps...", baudrate);
#endif

    switch(baudrate)
	{
		case 600:
			bps = B600;
			break;
		case 1200:
			bps = B1200;
			break;
		case 2400:
			bps = B2400;
			break;
		case 4800:
			bps = B4800;
			break;
		case 9600:
			bps = B9600;
			break;
		case 19200:
			bps = B19200;
			break;
	}
	cfsetispeed(&options, bps);
    cfsetospeed(&options, bps);

#ifdef VERBOSE
	printf("ok.\n");
#endif

    /* Set the new options for the port... */
    tcsetattr(fd, TCSANOW, &options);

	mdlns = 0;
	if (-1 == ioctl(fd, TIOCMGET, &mdlns))
	{
#ifdef VERBOSE
		printf("ERROR!\n");
#endif
		return (-1);
	}

	mdlns &= ~TIOCM_RTS;
	if (-1 == ioctl( fd, TIOCMSET, &mdlns )) 
  	{
	    close(fd);
	    fd = -1;

#ifdef VERBOSE
	    printf("ERROR!\n");
#endif
   	}
    return fd;
}

/* 
 * Reads a value from the DMM as a String.
 * "buffer" has a maximum of DMM_DATA_LENGTH_MAX characters
 * (e.g. 14 chars for a M-3650D. Error messages may be longer).
 */
void dmm_read(int fd, DMM *dmm)
{
	int n;

	for (n = 0; n < DMM_DATA_LENGTH_MAX; n++)
		dmm->rawdata[n] = '\0';

	/* When sending a "D" to the DMM, it sends a measured value back: */
	if (write(fd, "D\n", 2) < 0)
		strcpy(dmm->rawdata, "Error writing to DMM.\0");

	if (read(fd, dmm->rawdata, 14) < 0)
		strcpy(dmm->rawdata, "Error reading from DMM.\0");

	/* Last character of the string is not a CR? That probably means, */
	/* that our DMM is turned off or low on battery: */
	if (dmm->rawdata[13] != '\r')
		strcpy(dmm->rawdata, "DMM turned off?\0");
	else
	{
		/* Save rawdata: */
		if (dmm->value_array_counter < STORE_MAX)
		{
#ifdef VERBOSE
			printf("dmmread: Allocating %d bytes...\n",
			       strlen(dmm->rawdata) * sizeof(char) + 1);
#endif
			dmm->rawdata_array[dmm->value_array_counter] =
				malloc(strlen(dmm->rawdata) * sizeof(char));

			strcpy(dmm->rawdata_array[dmm->value_array_counter], dmm->rawdata);
			dmm->value_array_counter++;
		}
	}
}

void dmm_print_all_rawdata(DMM *dmm)
{
	int i;
	for (i = 0; i < dmm->value_array_counter; i++)
		printf("dmmread: [%d] %s \ndecoded: %f\n",
		       i, dmm->rawdata_array[i], dmm_decode_string_m3650(dmm, i));
}

int dmm_save_measurement_to_file(DMM *dmm, char *filename)
{
	FILE *file;
	int  i, result = FALSE;

	file = fopen(filename, "w");
	if (file)
	{
		fwrite("wejpdmm\n", 8 * sizeof(char), 1, file); /* File header */
		for (i = 0; i < dmm->value_array_counter; i++)
		{
#ifdef VERBOSE
			printf("dmmread: string groesse: %d\n", 
			       strlen(dmm->rawdata_array[i]) * sizeof(char));
#endif
			fwrite(dmm->rawdata_array[i], 
			       strlen(dmm->rawdata_array[i]) * sizeof(char), 
			       1,
			       file);
			fputc('\n', file);
		}
		fclose(file);
#ifdef VERBOSE
		printf("dmmread: %d values written to file %s.\n", i, filename);
#endif
		result = TRUE; /* Writing was succesful. */
	}
	
	return result;
}

int dmm_load_measurement_from_file(DMM *dmm, char *filename)
{
	FILE *file;
	int  i = 0, result = FALSE;
	char ch = 0, tmp_string[DMM_DATA_LENGTH_MAX] = "";

	dmm_init_struct(dmm);

	file = fopen(filename, "r");
	if (file)
	{
		fread(tmp_string, 8 * sizeof(char), 1, file);
		/* File header check: Is our file a wejpdmm file? */
		if (strcmp(tmp_string, "wejpdmm\n") == 0)
		{
			while (!feof(file))
			{
				for (i = 0; i < DMM_DATA_LENGTH_MAX; i++)
					tmp_string[i] = '\0';
				i = 0;
				ch = 0;
				while (ch != '\n' && !feof(file))
				{
					ch = fgetc(file);
					if (ch != '\n' && i < DMM_DATA_LENGTH_MAX)
					{
						tmp_string[i] = ch;
						i++;
					}
				}

				if (strlen(tmp_string) > 1)
				{
#ifdef VERBOSE
					printf("dmmread: %d. Allocating %d bytes...\n", 
					       dmm->value_array_counter,
					       (strlen(tmp_string) + 1) * sizeof(char));
#endif
					dmm->rawdata_array[dmm->value_array_counter] = 
						malloc((strlen(tmp_string) + 1) * sizeof(char));
#ifdef VERBOSE
					printf("dmmread: Saving data... \n%s\n", tmp_string);
#endif
					strncpy(dmm->rawdata_array[dmm->value_array_counter], 
					        tmp_string, strlen(tmp_string) * sizeof(char));

					/* String has to be zero terminiated! */
					*(dmm->rawdata_array[dmm->value_array_counter] +
						strlen(tmp_string)) = '\0';

					dmm->value_array_counter++;
				}
			}
			/* Calculate minimum and maximum: */
			for (i = 0; i < dmm->value_array_counter; i++)
				dmm_decode_string_m3650(dmm, i);
			result = TRUE; /* Reading was successful. */
		}
#ifdef VERBOSE
		printf("dmmread: Loading file %s %s\n",
		       filename, (result ? "successful." : "failed."));
#endif
		fclose(file);
	}
	return result;
}

/*
 * Calculates a float value from the string returned by the DMM. "record_number"
 * is the recorded string to be decoded.
 */
double dmm_decode_string_m3650(DMM *dmm, int record_number)
{
	short int sign = +1;
	float     factor = 1;
	int       n, number = 0, m, divisor = 1;
	float     tmp = 0.0;
	char      tmp_unit[32] = "";

	if (record_number >= dmm->value_array_counter || record_number < 0)
		record_number = dmm->value_array_counter;

	dmm->value = 0.0f;

	/* The last character from the DMM is always a CR: */
	if (*(dmm->rawdata_array[record_number] + 13) == '\r' &&
	    strlen(dmm->rawdata_array[record_number]) == 14)
	{
		m = 0;
		for (n = 9; n < 13; n++)
		{
			char ch = *(dmm->rawdata_array[record_number] + n);
			if (ch > ' ' && m < 5) /* Character "ch" is something printable? */
			{
				/* micro, milli etc. does not belong to the "unit", */
				/* ut changes the "factor": */
				if (m == 0)
				{
					if (ch == 'u') factor = 1.0F / (1000 * 1000); /* micro */
					if (ch == 'm') factor = 1.0F / 1000;          /* milli */
					if (ch == 'K') factor = 1000;                 /* kilo  */
					if (ch == 'M') factor = 1000 * 1000;          /* Mega  */
					if (ch == 'u' || ch == 'm' || ch == 'K' || ch == 'M') m--;
				}
				tmp_unit[m] = ch;
				m++;
			}
		}
		tmp_unit[m] = '\0';

#ifdef VERBOSE
		printf("dmmread: factor=%f \nunit=[%s]\n", factor, tmp_unit);
#endif

		/* If no unit has been saved yet, do it now: */
		if (strlen(dmm->unit) == 0) strncpy(dmm->unit, tmp_unit, 31);

		/* If unit has changed set the variable "unit_changed" to TRUE: */
		if (strcmp(dmm->unit, tmp_unit) != 0) dmm->unit_changed = TRUE;

		if (*(dmm->rawdata_array[record_number] + 3) == '-') sign = -1;

		m = 4;
		for (n = 4; n < 9; n++)
		{
			number = 0;
			switch(*(dmm->rawdata_array[record_number] + n))
			{
				case '1':
					number = 1; m--;
					break;
				case '2':
					number = 2; m--;
					break;
				case '3':
					number = 3; m--;
					break;
				case '4':
					number = 4; m--;
					break;
				case '5':
					number = 5; m--;
					break;
				case '6':
					number = 6; m--;
					break;
				case '7':
					number = 7; m--;
					break;
				case '8':
					number = 8; m--;
					break;
				case '9':
					number = 9; m--;
					break;
				case '0':
					number = 0; m--;
					break;
				case '.':
					divisor = pow(10.0, (8 - n));
					break;
			}

			tmp += pow(10.0, (float)m) * number;
		}
		dmm->value = sign * (tmp / divisor) * factor;

		/* New minimum or maximum value? */
		if (dmm->value > dmm->value_max) dmm->value_max = dmm->value;
		if (dmm->value < dmm->value_min) dmm->value_min = dmm->value;
	}
	return dmm->value;
}
