/*
 * gDMM Digital Multimeter Application
 *
 * File: graphwidget.c
 *
 * Copyright (c) 2003-2004 Johannes Heimansberg
 *
 * gDMM is a GTK+ 2 based application which reads data from a Digital Multimeter
 * (Metex/Voltcraft or compatible, e.g. M-3650D) through the serial interface
 * (RS-232) and draws a graph out of this data. Measurements can also be 
 * saved to/loaded from disk.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib.h>
#include "dmmread.h"
#include "graphwidget.h"

GtkWidget *graph_widget_create(GraphWidget *gw)
{
	gw->box = gtk_vbox_new(FALSE, 0);

	gw->frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(gw->frame), GTK_SHADOW_IN);
	gtk_container_add(GTK_CONTAINER(gw->box), gw->frame);

	gw->drawing_area = gtk_drawing_area_new();
	gtk_widget_set_size_request(gw->drawing_area, 240, 240);
	g_signal_connect(G_OBJECT(gw->drawing_area), "expose_event",    G_CALLBACK(graph_expose_event_callback),    gw);
	g_signal_connect(G_OBJECT(gw->drawing_area), "configure_event", G_CALLBACK(graph_configure_event_callback), gw);
	gtk_container_add(GTK_CONTAINER(gw->frame), gw->drawing_area);

	gw->adjustment = (GtkAdjustment*)gtk_adjustment_new(0, 0, 100, 1, 10, 10);
	gw->scrollbar = gtk_hscrollbar_new(gw->adjustment);
	g_signal_connect(G_OBJECT(gw->adjustment), "value_changed", G_CALLBACK(graph_scrollbar_moved_callback), gw);
	gtk_box_pack_start(GTK_BOX(gw->box), gw->scrollbar, FALSE, TRUE, 0);

	gtk_widget_show(gw->box);

	gw->ZOOM_DEFAULT = 6;
	gw->zoom = gw->ZOOM_DEFAULT;

	gw->colormap = gdk_colormap_new(gdk_visual_get_best(), FALSE);

	gw->pgl = pango_layout_new(gdk_pango_context_get_for_screen(gdk_screen_get_default()));

	return gw->box;
}

void graph_zoom_in(GraphWidget *gw)
{
	if (gw->zoom < 20) gw->zoom++;
}

void graph_zoom_out(GraphWidget *gw)
{
	if (gw->zoom > 1) gw->zoom--;
}

void graph_zoom_default(GraphWidget *gw)
{
	gw->zoom = gw->ZOOM_DEFAULT;
}

void graph_draw(GraphWidget *gw)
{
	long   max_count, i, offset;
	double max = 0.0, distance;

	GdkColor white = { 0, 65535, 65535, 65535, };
	GdkColor grey  = { 0, 32000, 32000, 32000, };
	GdkColor black = { 0,     0,     0,     0, };
	GdkColor green = { 0,     0, 65535,     0, };

	gdk_rgb_find_color(gw->colormap, &white);
	gdk_rgb_find_color(gw->colormap, &black);
	gdk_rgb_find_color(gw->colormap, &grey);
	gdk_rgb_find_color(gw->colormap, &green);

	gdk_gc_set_background(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &black);
  	gdk_window_set_background (gw->drawing_area->window, &black);
	gdk_window_show (gw->drawing_area->window);

	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &black);
	gdk_draw_rectangle((GdkDrawable*)gw->pixmap,
	                   gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], TRUE,
	                   0, 0, gw->drawing_area->allocation.width, gw->drawing_area->allocation.height);

	/* Draw zero line: */
	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &grey);
	gdk_draw_line((GdkDrawable*)gw->pixmap, gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
	              0, gw->drawing_area->allocation.height / 2,
	              gw->drawing_area->allocation.width, gw->drawing_area->allocation.height / 2);

	/* Calculate absolute maximum: */
	max = dmm_get_max_value(&(gw->dmm));
	if (max < -dmm_get_min_value(&(gw->dmm))) max = -dmm_get_min_value(&(gw->dmm));

	/* Horizontal lines with lettering: */
	distance = max / 4.0;
	for (i = -4; i <= 4; i++)
	{
		char  tmp[512];
		float tmpvalue = distance * i;
		int   y = gw->drawing_area->allocation.height / 2 - (tmpvalue / max) * (gw->drawing_area->allocation.height / 2);

		gdk_draw_line((GdkDrawable*)gw->pixmap,
		              gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
		              0, y, gw->drawing_area->allocation.width, y);

		sprintf(tmp, "%3.3f", tmpvalue);	
		pango_layout_set_text(gw->pgl, tmp, strlen(tmp));
		gdk_draw_layout((GdkDrawable*)gw->pixmap,
		                gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], 10, y, gw->pgl);
	}

	/* Show current measurement value: */
	pango_layout_set_text(gw->pgl, dmm_get_rawdata(&(gw->dmm)), strlen(dmm_get_rawdata(&(gw->dmm))));
	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &green);
	gdk_draw_layout((GdkDrawable*)gw->pixmap, gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
	                25, 10, gw->pgl);

	/* Show unit: */
	pango_layout_set_text(gw->pgl, dmm_get_unit(&(gw->dmm)), strlen(dmm_get_unit(&(gw->dmm))));
	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &green);
	gdk_draw_layout((GdkDrawable*)gw->pixmap, gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
	                gw->drawing_area->allocation.width - 60, 10, gw->pgl);

	/* Get number of lines to draw (depends on the number of measured values and/or of the widget's width): */
	max_count = (dmm_get_number_of_stored_values(&(gw->dmm)) > gw->drawing_area->allocation.width ? 
	            gw->drawing_area->allocation.width : dmm_get_number_of_stored_values(&(gw->dmm)) - 1);

	offset = (int)gtk_adjustment_get_value(gw->adjustment);
	
	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &grey);
	/* Draw time markings: */
	for (i = 0; i <= max_count / 10; i++)
	{
		char tmp[512];

		if (i * 10 - offset % 10 > 0)
		{
			gdk_draw_line((GdkDrawable*)gw->pixmap,
			              gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
			              (i * 10 - offset % 10) * gw->zoom, 0, 
			              (i * 10 - offset % 10) * gw->zoom, gw->drawing_area->allocation.height);

			if (gw->zoom > 3)
			{
				sprintf(tmp, "%ld", i + offset / 10);
				pango_layout_set_text(gw->pgl, tmp, strlen(tmp));
				gdk_draw_layout((GdkDrawable*)gw->pixmap,
			    	            gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
			    	            (i * 10 - offset % 10) * gw->zoom + 1,
			    	            gw->drawing_area->allocation.height - 20,
			    	            gw->pgl);
			}
		}
	}

	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &green);
	/* Draw the graph: */
	for (i = 0; i < max_count; i++)
	{
		/* Draw the graph with optimal vertical expanse: */
		gdk_draw_line((GdkDrawable*)gw->pixmap, 
		              gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)],
		              i * gw->zoom,
		              gw->drawing_area->allocation.height / 2 - (int)((dmm_get_stored_value(&(gw->dmm), offset + i) / max) * gw->drawing_area->allocation.height / 2),
		              (i+1) * gw->zoom,
		              gw->drawing_area->allocation.height / 2 - (int)((dmm_get_stored_value(&(gw->dmm), offset + i + 1) / max) * gw->drawing_area->allocation.height / 2));
	}

	gdk_gc_set_foreground(gw->drawing_area->style->fg_gc[GTK_WIDGET_STATE(gw->drawing_area)], &black);

	GdkRectangle rect;
	rect.x = 0;
	rect.y = 0;
	rect.width = gw->drawing_area->allocation.width;
	rect.height = gw->drawing_area->allocation.height;
	gtk_widget_draw(gw->drawing_area, &rect);
}

void graph_update_adjustment(GraphWidget *gw)
{
	gw->adjustment->page_size = gw->drawing_area->allocation.width / gw->zoom;

	gw->adjustment->lower = 0;
	gw->adjustment->upper = (gw->drawing_area->allocation.width / gw->zoom >= 
	                         (dmm_get_number_of_stored_values(&(gw->dmm))) ?
	                         1 : dmm_get_number_of_stored_values(&(gw->dmm)));

	gtk_adjustment_changed(gw->adjustment);
}

/******* callback functions *******/
gboolean graph_configure_event_callback(GtkWidget *widget, GdkEventConfigure *event, GraphWidget *gw)
{
	if (gw->pixmap)
		gdk_pixmap_unref((GdkDrawable*)gw->pixmap);

	gw->pixmap = (GtkPixmap*)gdk_pixmap_new(widget->window,
	                        widget->allocation.width,
	                        widget->allocation.height,
	                        -1);
	gdk_draw_rectangle((GdkDrawable*)gw->pixmap,
	                   widget->style->black_gc,
	                   TRUE,
	                   0, 0,
	                   widget->allocation.width,
	                   widget->allocation.height);
	graph_draw(gw);

	graph_update_adjustment(gw);
	gtk_adjustment_clamp_page(gw->adjustment, gw->adjustment->lower, gw->adjustment->upper);
	return TRUE;
}

gboolean graph_expose_event_callback(GtkWidget *widget, GdkEventExpose *event, GraphWidget *gw)
{
	gdk_draw_pixmap(widget->window,
	                widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
	                (GdkDrawable*)gw->pixmap, 0, 0, 0, 0, widget->allocation.width, widget->allocation.height);
/*	                event->area.x, event->area.y,
	                event->area.x, event->area.y,
	                event->area.width, event->area.height); */
	graph_update_adjustment(gw);
	return TRUE;
}

void graph_scrollbar_moved_callback(GtkAdjustment *adjustment, GraphWidget *gw)
{
	graph_draw(gw);
	gtk_widget_queue_draw_area(gw->drawing_area,
	                           0, 0,
	                           gw->drawing_area->allocation.width, gw->drawing_area->allocation.height);
}
/**********************************/
