/*
 * gDMM Digital Multimeter Application
 *
 * File: main.c
 *
 * Copyright (c) 2003-2004 Johannes Heimansberg
 *
 * gDMM is a GTK+ 2 based application which reads data from a Digital Multimeter
 * (Metex/Voltcraft or compatible, e.g. M-3650D) through the serial interface
 * (RS-232) and draws a graph out of this data. Measurements can also be 
 * saved to/loaded from disk.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define VERSION_NUMBER "0.1.2"
/* #define VERBOSE 1 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <termio.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <math.h>
#include <pthread.h>
#include <sys/time.h>
#include "dmmread.h"
#include "graphwidget.h"
#include "main.h"
#include "wejpconfig.h"
#include "configdialog.h"
#include "gdmm_mini.xpm"

static pthread_t     dmm_read_thread;
static int           dmm_read_thread_running;
GtkWidget            *window;    /* Main window */
static GtkWidget     *statusbar; /* Main window's status bar */
static GraphWidget   gw;
char                 *path_to_config_file;

/* This thread reads the data from the DMM: */
static void *dmm_read_thread_loop(void *parameter)
{
	int             fd;
	char            device[128];
	ConfigFile      cf;
	struct timeval  timer_start, timer_end;
	struct timespec req, rem;

	dmm_read_thread_running = TRUE;

	/* Get selected com port device from config file: */
	cfg_read_config_file(&cf, path_to_config_file);
	strncpy(device, cfg_get_key_value(cf, "Port"), 127);
	if (device == NULL) sprintf(device, "/dev/ttyS0");
	cfg_free_config_file_struct(&cf);

#ifdef VERBOSE
	printf("main: Opening serial port (%s)...", device);
#endif

	fd = dmm_open_port(device, 1200, 7, 2);

#ifdef VERBOSE
	if (fd > 0)
		printf("ok.\n");
	else
		printf("failed!\n");
#endif

	/* If opening serial device was not successful show an error message dialog: */
	if (fd < 0)
	{
		GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                        	GTK_DIALOG_DESTROY_WITH_PARENT,
                        	GTK_MESSAGE_ERROR,
                        	GTK_BUTTONS_OK,
                        	"Cannot open serial device %s !",
                        	device);
		gdk_threads_enter();
		gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);
		gdk_threads_leave();
	}
	
	while (dmm_read_thread_running == TRUE && fd > 0)
	{
		gettimeofday(&timer_start, NULL);
		timer_end = timer_start;

		dmm_read(fd, &(gw.dmm));
		/* If the unit changes during measurement (that should not happen!), */
		/* abort measurement: */
		if (dmm_is_unit_changed(&(gw.dmm)) == TRUE)
			dmm_read_thread_running = FALSE;
		else
		{

#ifdef VERBOSE
			printf("\nmain: Data from DMM: %s\n", dmm_get_rawdata(&(gw.dmm)));
			printf("main: Decoded: Value=%f Unit=%s\n\n",
			       dmm_get_last_value(&(gw.dmm)), dmm_get_unit(&(gw.dmm)));
#endif

			gdk_threads_enter();

			/* Always move the scrollbar to the most right, so the newest */
			/* part of the graph is always visible: */
			gw.adjustment->value = (gw.adjustment->upper - gw.adjustment->page_size > 0 ?
		    	                   gw.adjustment->upper - gw.adjustment->page_size : 0);
			gtk_adjustment_value_changed(gw.adjustment);

			gdk_threads_leave();

			gettimeofday(&timer_end, NULL);
			req.tv_sec = 0;
			/* Each loop cycle lasts exactly 1 second: */
			req.tv_nsec = (1000000 - (timer_end.tv_usec - timer_start.tv_usec)) * 1000;

			/* If nanosleep was interrupted, retry with the time remaining: */
			while (nanosleep(&req, &rem) < 0)
			{   /* until sleeping was successful. */
				req.tv_nsec = rem.tv_nsec;
			}
		}
	}

#ifdef VERBOSE
	printf("main: Closing serial port...");
#endif

	close(fd);

#ifdef VERBOSE
	printf("ok.\n\n");
#endif

	dmm_read_thread_running = FALSE;

	pthread_join(dmm_read_thread, NULL);
	return NULL;
}

GtkWidget *create_menu(GtkWidget *window)
{
	static GtkWidget *menubar;
	static GtkWidget *file_menu, *file_item;
	static GtkWidget *new_item, *open_item, *save_item, *start_item;
	static GtkWidget *stop_item, *preferences_item, *quit_item;
	static GtkWidget *separator1_item, *separator2_item, *separator3_item;

	static GtkWidget *view_menu, *view_item;
	static GtkWidget *zoom_in_item, *zoom_out_item, *zoom_100_item;

	static GtkWidget *help_menu, *help_item;
	static GtkWidget *about_item;

	static GtkAccelGroup *accel_group;
	
	accel_group = gtk_accel_group_new();

	/* File menu */
	file_menu = gtk_menu_new();

	new_item         = gtk_image_menu_item_new_from_stock("gtk-new", NULL);	
	open_item        = gtk_image_menu_item_new_from_stock("gtk-open", NULL);
	save_item        = gtk_image_menu_item_new_from_stock("gtk-save-as", NULL);
	start_item       = gtk_menu_item_new_with_mnemonic("St_art");
	stop_item        = gtk_image_menu_item_new_from_stock("gtk-stop", NULL);
	preferences_item = gtk_image_menu_item_new_from_stock("gtk-preferences", NULL);
	quit_item        = gtk_image_menu_item_new_from_stock("gtk-quit", NULL);
	separator1_item  = gtk_separator_menu_item_new();
	separator2_item  = gtk_separator_menu_item_new();
	separator3_item  = gtk_separator_menu_item_new();

	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), new_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), open_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), save_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), separator1_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), start_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), stop_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), separator2_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), preferences_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), separator3_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), quit_item);

	gtk_widget_show(new_item);
	gtk_widget_show(open_item);
	gtk_widget_show(save_item);
	gtk_widget_show(preferences_item);
	gtk_widget_show(quit_item);

	g_signal_connect(G_OBJECT(new_item), "activate",
                             G_CALLBACK(reset),
                             &gw);

	g_signal_connect_swapped(G_OBJECT(quit_item), "activate",
                             G_CALLBACK(quit_function),
                             (gpointer) "file.quit");
                             
    g_signal_connect_swapped(G_OBJECT(open_item), "activate",
                             G_CALLBACK(open_function),
                             (gpointer) "file.open");

    g_signal_connect_swapped(G_OBJECT(save_item), "activate",
                             G_CALLBACK(save_as_function),
                             (gpointer) "file.save_as");

	g_signal_connect_swapped(G_OBJECT(start_item), "activate",
	                         G_CALLBACK(start_measurement),
	                         (gpointer) "file.start");

	g_signal_connect_swapped(G_OBJECT(stop_item), "activate",
	                         G_CALLBACK(stop_measurement),
	                         (gpointer) "file.stop");

	g_signal_connect_swapped(G_OBJECT(preferences_item), "activate",
	                         G_CALLBACK(preferences_dialog),
	                         (gpointer) "file.preferences");

	gtk_widget_add_accelerator(new_item,
	                           "activate",
	                           accel_group,
	                           'N',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(quit_item,
	                           "activate",
	                           accel_group,
	                           'Q',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(open_item,
	                           "activate",
	                           accel_group,
	                           'O',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(save_item,
	                           "activate",
	                           accel_group,
	                           'S',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(start_item,
	                           "activate",
	                           accel_group,
	                           'A',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(stop_item,
	                           "activate",
	                           accel_group,
	                           'T',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);

	/* View menu */
	view_menu = gtk_menu_new();
    
	zoom_in_item  = gtk_image_menu_item_new_from_stock("gtk-zoom-in",  NULL);
	zoom_out_item = gtk_image_menu_item_new_from_stock("gtk-zoom-out", NULL);
	zoom_100_item = gtk_image_menu_item_new_from_stock("gtk-zoom-100", NULL);

	gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), zoom_in_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), zoom_out_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), zoom_100_item);

	g_signal_connect(G_OBJECT(zoom_in_item), "activate",
	                 G_CALLBACK(zoom_in),
	                 &gw);

	g_signal_connect(G_OBJECT(zoom_out_item), "activate",
	                 G_CALLBACK(zoom_out),
	                 &gw);

	g_signal_connect(G_OBJECT(zoom_100_item), "activate",
	                 G_CALLBACK(zoom_default),
	                 &gw);

	gtk_widget_add_accelerator(zoom_in_item,
	                           "activate",
	                           accel_group,
	                           '+',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(zoom_out_item,
	                           "activate",
	                           accel_group,
	                           '-',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);
	gtk_widget_add_accelerator(zoom_100_item,
	                           "activate",
	                           accel_group,
	                           '1',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);

	/* Help menu */
	help_menu = gtk_menu_new();

	about_item = gtk_menu_item_new_with_mnemonic("_About");

	gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), about_item);

	g_signal_connect_swapped(G_OBJECT(about_item), "activate",
                             G_CALLBACK(about_function),
                             (gpointer) "help.about");

	gtk_widget_add_accelerator(about_item,
	                           "activate",
	                           accel_group,
	                           'H',
	                           GDK_CONTROL_MASK,
	                           GTK_ACCEL_VISIBLE);

	/* Menu bar */
	menubar = gtk_menu_bar_new();

	file_item = gtk_menu_item_new_with_mnemonic("_File");
	gtk_widget_show(file_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_item), file_menu);
	gtk_menu_bar_append(GTK_MENU_BAR(menubar), file_item);

	view_item = gtk_menu_item_new_with_mnemonic("_View");
	gtk_widget_show(view_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(view_item), view_menu);
	gtk_menu_bar_append(GTK_MENU_BAR(menubar), view_item);

	help_item = gtk_menu_item_new_with_mnemonic("_Help");
	gtk_widget_show(help_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_item), help_menu);
	gtk_menu_bar_append(GTK_MENU_BAR(menubar), help_item);

	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

	return menubar;
}

/*********** Callback functions ************/
void start_measurement(GtkWidget *widget, gpointer data)
{
	if (dmm_read_thread_running == FALSE)
	{
#ifdef VERBOSE
		g_print("main: Starting measurement...\n");
#endif
		pthread_create(&dmm_read_thread, NULL, dmm_read_thread_loop, NULL);
		gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
		                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                               "nothing"));
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	   	   	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
		                   "Measurement started.");
	}
}

void stop_measurement(GtkWidget *widget, gpointer data)
{
	if (dmm_read_thread_running)
	{
#ifdef VERBOSE
		g_print("main: Aborting measurement...\n");
#endif
		dmm_read_thread_running = FALSE;
		gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
		                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                               "nothing"));
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	   	   	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
		                   "Measurement stopped.");
	}
}

/* callback: close window (return FALSE = close window) */
gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	dmm_read_thread_running = FALSE;
#ifdef VERBOSE
	g_print("main: Exit request!\n");
#endif
	return FALSE;
}

/* callback: exit application */
void quit_function(GtkWidget *widget, gpointer data)
{
	dmm_read_thread_running = FALSE;
#ifdef VERBOSE
	g_print("main: Exit...\n");
#endif
	gtk_main_quit();
}

void reset(GtkWidget *widget, GraphWidget *gwt)
{
	dmm_init_struct(&(gwt->dmm));
	graph_draw(gwt);
	gtk_widget_queue_draw_area(gwt->drawing_area,
	                           0,
	                           0,
	                           gwt->drawing_area->allocation.width,
	                           gwt->drawing_area->allocation.height);
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
	                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
	                                               "nothing"));
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	      	           gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
	                                                "nothing"),
	        	       "All cleared.");
}

void save_file(GtkWidget *widget, GtkFileSelection *fs)
{
	dmm_read_thread_running = FALSE;
	gtk_widget_hide(GTK_WIDGET(fs));
	if ( dmm_save_measurement_to_file(
		    &(gw.dmm),
		    (char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs))
	     ) 
	   )
	{
		gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
		                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                               "nothing"));
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	      	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
	        	           "File saved.");
	}
	else
	{
		gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
		                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                               "nothing"));
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	      	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
	        	           "Failed writing to file.");
	}
}

void load_file(GtkWidget *widget, GtkFileSelection *fs)
{
	int success;
	dmm_read_thread_running = FALSE;
	gtk_widget_hide(GTK_WIDGET(fs));
	success = 
		dmm_load_measurement_from_file(
			&(gw.dmm),
	    	(char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs))
		);

	gw.adjustment->value = 0; /* Move scrollbar to the left */
	graph_draw(&gw);
	gtk_widget_queue_draw_area(gw.drawing_area,
	                           0,
	                           0,
	                           gw.drawing_area->allocation.width,
	                           gw.drawing_area->allocation.height);

	gtk_statusbar_pop(GTK_STATUSBAR(statusbar),
	                  gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
	                                               "nothing"));
	if (success)
	{
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	      	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
	        	           "File loaded.");
	}
	else
	{
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
	      	               gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		                                                "nothing"),
	        	           "File loading failed.");
	}
}

void close_file_dialog(GtkWidget *widget, gpointer data)
{
	gtk_widget_hide(GTK_WIDGET(data));
}

void open_function(GtkWidget *widget, gpointer data)
{
	GtkWidget *filew;
	dmm_read_thread_running = FALSE;
	filew = gtk_file_selection_new("Open...");
	gtk_window_set_transient_for(GTK_WINDOW(filew), GTK_WINDOW(window));
	g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(filew)->ok_button),
	                 "clicked",
	                 G_CALLBACK(load_file),
	                 (gpointer)filew);
	g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button),
	                 "clicked",
	                 G_CALLBACK(close_file_dialog),
	                 (gpointer)filew);
	gtk_file_selection_complete(GTK_FILE_SELECTION(filew), "*.dmm");
	gtk_widget_show(filew);
}

void save_as_function(GtkWidget *widget, gpointer data)
{
	GtkWidget *filew;
	dmm_read_thread_running = FALSE;
	filew = gtk_file_selection_new("Save as...");
	gtk_window_set_transient_for(GTK_WINDOW(filew), GTK_WINDOW(window));

	g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(filew)->ok_button),
	                 "clicked",
	                 G_CALLBACK(save_file),
	                 (gpointer)filew);
	g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button),
	                 "clicked",
	                 G_CALLBACK(close_file_dialog),
	                 (gpointer)filew);
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(filew), "measurement1.dmm");
	gtk_widget_show(filew);
}

/* Program info */
void about_function(GtkWidget *widget, gpointer data)
{
	GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                        	GTK_DIALOG_DESTROY_WITH_PARENT,
                        	GTK_MESSAGE_INFO,
                        	GTK_BUTTONS_OK,
                        	"gDMM %s\n\nCopyright (c) 2003-2004 Johannes " \
	                        "Heimansberg\n\n" \
	                        "http://www.wejp.de/\n\nThis program is free " \
	                        "software released under the terms of the GPL " \
	                        "(see COPYING).\n\n" \
							"This program is distributed in the hope that it" \
	                        " will be useful, but WITHOUT ANY WARRANTY; " \
	                        " without even the implied warranty of " \
	                        "MERCHANTABILITY or FITNESS FOR A PARTICULAR " \
	                        "PURPOSE. See the GNU General Public License " \
	                        "for more details.",
                        	VERSION_NUMBER);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

void preferences_dialog(GtkWidget *widget, gpointer data)
{
	static SettingsDialog sd;
	sd.parent = window;
	config_dialog_show(&sd);
}

void zoom_in(GtkWidget *widget, GraphWidget *gw)
{
	graph_zoom_in(gw);
	graph_draw(gw);
	if (gw->adjustment->value > gw->adjustment->upper - gw->adjustment->page_size)
		gw->adjustment->value = (gw->adjustment->upper - gw->adjustment->page_size < 0 ?
		                         0 : gw->adjustment->upper - gw->adjustment->page_size);
	gtk_adjustment_value_changed(gw->adjustment);
	gtk_widget_queue_draw_area(gw->drawing_area,
	                           0,
	                           0,
	                           gw->drawing_area->allocation.width,
	                           gw->drawing_area->allocation.height);

}

void zoom_out(GtkWidget *widget, GraphWidget *gw)
{
	graph_zoom_out(gw);
	graph_draw(gw);
	if (gw->adjustment->value > gw->adjustment->upper - gw->adjustment->page_size)
		gw->adjustment->value = (gw->adjustment->upper - gw->adjustment->page_size < 0 ?
		                         0 : gw->adjustment->upper - gw->adjustment->page_size);
	gtk_adjustment_value_changed(gw->adjustment);
	gtk_widget_queue_draw_area(gw->drawing_area,
	                           0,
	                           0,
	                           gw->drawing_area->allocation.width,
	                           gw->drawing_area->allocation.height);
}

void zoom_default(GtkWidget *widget, GraphWidget *gw)
{
	graph_zoom_default(gw);
	graph_draw(gw);
	if (gw->adjustment->value > gw->adjustment->upper - gw->adjustment->page_size)
		gw->adjustment->value = (gw->adjustment->upper - gw->adjustment->page_size < 0 ?
		                         0 : gw->adjustment->upper - gw->adjustment->page_size);
	gtk_adjustment_value_changed(gw->adjustment);
	gtk_widget_queue_draw_area(gw->drawing_area,
	                           0,
	                           0,
	                           gw->drawing_area->allocation.width,
	                           gw->drawing_area->allocation.height);
}
/******************************/

int main(int argc, char *argv[])
{
	GtkWidget *mainbox;
	GtkWidget *button_exit;
	GtkWidget *button_start, *button_stop;
	GtkWidget *button_reset;
	GtkWidget *button_zoom_in, *button_zoom_out;
	GtkWidget *menubar;
	GtkWidget *toolbar;
	GdkPixbuf *icon;
	char      *home_directory;

	dmm_init_struct(&(gw.dmm));

	/* Do we have an command line argument? */
	if (argc > 1)
	{
		/* The comand line argument has to be a .dmm file. Try to load it: */
		dmm_load_measurement_from_file(&(gw.dmm), argv[1]);
	}

	/* Set path to config file to ~/.gdmmrc */	
	home_directory = getenv("HOME");
	path_to_config_file = (char*) malloc(strlen(home_directory) + 128);
	sprintf(path_to_config_file, "%s/.gdmmrc", home_directory);

	if (cfg_check_config_file(path_to_config_file) < 0)
	{
		ConfigFile cf;
		cfg_init_config_file_struct(&cf);
		printf("No config file found! Creating default config file...\n");
		cfg_add_key(&cf, "DMMClass", "Metex_RS232");
		cfg_add_key(&cf, "DMMModel", "Metex_M-3xxxD");
		cfg_add_key(&cf, "Port", "/dev/ttyS0");
		cfg_write_config_file(&cf, path_to_config_file);
	}

	g_thread_init(NULL);
	gdk_threads_init();
	gtk_init(&argc, &argv);

	mainbox = gtk_vbox_new(FALSE, 3);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	
	icon = gdk_pixbuf_new_from_xpm_data((const char **)gdmm_mini_xpm);
	gtk_window_set_icon (GTK_WINDOW (window), icon);
	
	gtk_container_set_border_width(GTK_CONTAINER(window), 0);

	gtk_window_set_policy(GTK_WINDOW(window), FALSE, TRUE, FALSE);

	gtk_window_set_default_size(GTK_WINDOW(window), 620, 460);

	gtk_window_set_title(GTK_WINDOW(window), "gDMM");

	g_signal_connect(G_OBJECT(window), "delete_event",
	                 G_CALLBACK(delete_event),  NULL);
	g_signal_connect(G_OBJECT(window), "destroy",
	                 G_CALLBACK(quit_function), NULL);

	menubar = create_menu(window);
	gtk_box_pack_start(GTK_BOX(mainbox), menubar, FALSE, TRUE, 0);

	toolbar = gtk_toolbar_new();
	gtk_box_pack_start(GTK_BOX(mainbox), toolbar, FALSE, TRUE, 0);

	button_start    = gtk_image_new_from_stock(GTK_STOCK_EXECUTE,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);
	button_stop     = gtk_image_new_from_stock(GTK_STOCK_STOP,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);
	button_reset    = gtk_image_new_from_stock(GTK_STOCK_CLEAR,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);
	button_exit     = gtk_image_new_from_stock(GTK_STOCK_QUIT,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);
	button_zoom_in  = gtk_image_new_from_stock(GTK_STOCK_ZOOM_IN,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);
	button_zoom_out = gtk_image_new_from_stock(GTK_STOCK_ZOOM_OUT,
	                                           GTK_ICON_SIZE_SMALL_TOOLBAR);

	gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
	                        "Start",
	                        "Start measurement",
	                        "private",
	                        button_start,
	                        G_CALLBACK(start_measurement),
	                        NULL);

	gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
	                        "Stop",
	                        "Stop measurement",
	                        "private",
	                        button_stop,
	                        G_CALLBACK(stop_measurement),
	                        NULL);

	gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
	                        "Reset",
	                        "Clear all",
	                        "private",
	                        button_reset,
	                        G_CALLBACK(reset),
	                        &gw);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
	                        "Zoom in",
	                        "Zoom in",
	                        "private",
	                        button_zoom_in,
	                        G_CALLBACK(zoom_in),
	                        &gw);

	gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
	                        "Zoom out",
	                        "Zoom out",
	                        "private",
	                        button_zoom_out,
	                        G_CALLBACK(zoom_out),
	                        &gw);

	gtk_widget_show(button_start);
	gtk_widget_show(button_stop);
	gtk_widget_show(button_reset);
	gtk_widget_show(button_exit);
	gtk_widget_show(button_zoom_in);
	gtk_widget_show(button_zoom_out);

	gtk_container_add(GTK_CONTAINER(mainbox), graph_widget_create(&gw));

	statusbar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(mainbox), statusbar, FALSE, TRUE, 0);
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), TRUE);
	gtk_widget_show(statusbar);
	gtk_statusbar_push(GTK_STATUSBAR(statusbar),
	                   gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
	                                                "nothing"), 
	                   "Ready.");

    gtk_container_add(GTK_CONTAINER(window), mainbox);

	gtk_widget_show_all(window);
	
	gdk_threads_enter();
	gtk_main();
	gdk_threads_leave();

	dmm_clear_struct(&(gw.dmm));

	free(path_to_config_file);

	return 0;
}
